/* -*-c-*-
 *
 * $Id: txport.c,v 1.1 2002/01/25 19:34:45 mdw Exp $
 *
 * Transport switch glue
 *
 * (c) 2001 Mark Wooding
 */

/*----- Licensing notice --------------------------------------------------* 
 *
 * This file is part of Jog: Programming for a jogging machine.
 *
 * Jog is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * Jog is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with Jog; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*----- Revision history --------------------------------------------------* 
 *
 * $Log: txport.c,v $
 * Revision 1.1  2002/01/25 19:34:45  mdw
 * Initial revision
 *
 */

/*----- Header files ------------------------------------------------------*/

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <errno.h>
#include <stdio.h>
#include <stdarg.h>

#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>

#include <mLib/darray.h>
#include <mLib/dstr.h>
#include <mLib/lbuf.h>
#include <mLib/sub.h>
#include <mLib/trace.h>
#include <mLib/tv.h>

#include "err.h"
#include "txport.h"

/*----- Global variables --------------------------------------------------*/

#define TX_LIST 0
#include "tx-socket.h"
#include "tx-serial-unix.h"
txport_ops *txlist = TX_LIST;

const char *txname = 0;
const char *txfile = 0;
const char *txconf = 0;

/*----- Main code ---------------------------------------------------------*/

/* --- @newline@ --- *
 *
 * Arguments:	@char *s@ = pointer to line
 *		@size_t len@ = length of line
 *		@void *txv@ = pointer to transport context
 *
 * Returns:	---
 *
 * Use:		Adds a line to the list.
 */

static void newline(char *s, size_t len, void *txv)
{
  txport *tx = txv;
  txline *l;

  if (!s)
    return;
  l = CREATE(txline);
  l->s = xmalloc(len + 1);
  memcpy(l->s, s, len + 1);
  l->len = len;
  l->tx = tx;
  l->next = 0;
  l->prev = tx->ll_tail;
  if (tx->ll_tail)
    tx->ll_tail->next = l;
  else
    tx->ll = l;
  tx->ll_tail = l;
}

/* --- @tx_create@ --- *
 *
 * Arguments:	@const char *name@ = name of transport to instantiate
 *		@const char *file@ = filename for transport
 *		@const char *config@ = config string
 *
 * Returns:	A pointer to the transport context, or null on error.
 *
 * Use:		Creates a new transport.
 */

txport *tx_create(const char *name, const char *file, const char *config)
{
  txport_ops *o;
  txport *tx;
  pthread_attr_t ta;
  dstr d = DSTR_INIT;
  size_t len;
  int e;

  /* --- Look up the transport by name --- */

  if (!name) {
    o = txlist;
    goto found;
  }
  len = strlen(name);
  for (o = txlist; o; o = o->next) {
    if (strncmp(name, o->name, len) == 0)
      goto found;
  }
  err_report(ERR_TXPORT, ERRTX_BADTX, 0, "unknown transport `%s'", name);
  return (0);

  /* --- Set up the transport block --- */

found:
  if (!file) {
    const struct txfile *fv;
    for (fv = o->fv; fv->env || fv->name; fv++) {
      if (fv->env && (file = getenv(fv->env)) == 0)
	continue;
      DRESET(&d);
      if (file)
	DPUTS(&d, file);
      if (file && fv->name)
	DPUTC(&d, '/');
      if (fv->name)
	DPUTS(&d, fv->name);
      break;
    }
    file = d.buf;
  }
  if (!config)
    config = o->config;
  if ((tx = o->create(file, config)) == 0)
    goto fail_0;
  tx->ops = o;
  DA_CREATE(&tx->buf);
  tx->ll = 0;
  tx->ll_tail = 0;
  if ((e = pthread_mutex_init(&tx->mx, 0)) != 0) {
    err_report(ERR_TXPORT, ERRTX_CREATE, e,
	       "mutex creation failed: %s", strerror(e));
    goto fail_1;
  }
  if ((e = pthread_cond_init(&tx->cv, 0)) != 0) {
    err_report(ERR_TXPORT, ERRTX_CREATE, e,
	       "condvar creation failed: %s", strerror(e));
    goto fail_2;
  }
  if ((e = pthread_attr_init(&ta)) != 0) {
    err_report(ERR_TXPORT, ERRTX_CREATE, e,
	       "thread attribute creation failed: %s", strerror(e));
    goto fail_3;
  } 
  if ((e = pthread_attr_setdetachstate(&ta, PTHREAD_CREATE_DETACHED)) ||
      (e = pthread_create(&tx->tid, &ta, tx->ops->fetch, tx)) != 0) {
    err_report(ERR_TXPORT, ERRTX_CREATE, e,
	       "thread creation failed: %s", strerror(e));
    goto fail_4;
  }
  pthread_attr_destroy(&ta);
  lbuf_init(&tx->lb, newline, tx);
  tx->lb.delim = '\r';
  tx->s = TX_READY;
  DDESTROY(&d);
  return (tx);

  /* --- Something went wrong --- */

fail_4:
  pthread_attr_destroy(&ta);
fail_3:
  pthread_cond_destroy(&tx->cv);
fail_2:
  pthread_mutex_destroy(&tx->mx);
fail_1:
  tx->ops->destroy(tx);
fail_0:
  DDESTROY(&d);
  return (0);
}

/* --- @tx_write@ --- *
 *
 * Arguments:	@txport *tx@ = pointer to transport context
 *		@const void *p@ = pointer to buffer to write
 *		@size_t sz@ = size of buffer
 *
 * Returns:	Zero if OK, or @-1@ on error.
 *
 * Use:		Writes some data to a transport.
 */

int tx_write(txport *tx, const void *p, size_t sz)
{
  if (tx->ops->write(tx, p, sz) < 0) {
    err_report(ERR_TXPORT, ERRTX_WRITE, errno,
	       "error writing to transport: %s", strerror(errno));
    return (-1);
  }
  return (0);
}

/* --- @tx_printf@ --- *
 *
 * Arguments:	@txport *tx@ = pointer to transport context
 *		@const char *p@ = pointer to string to write
 *
 * Returns:	The number of characters printed, or @EOF@ on error.
 *
 * Use:		Writes a textual message to a transport.
 */

int tx_vprintf(txport *tx, const char *p, va_list *ap)
{
  dstr d = DSTR_INIT;
  int rc;

  dstr_vputf(&d, p, *ap);
  rc = d.len;
  rc = tx_write(tx, d.buf, d.len);
  DDESTROY(&d);
  return (rc);
}

int tx_printf(txport *tx, const char *p, ...)
{
  va_list ap;
  int rc;

  va_start(ap, p);
  rc = tx_vprintf(tx, p, &ap);
  va_end(ap);
  return (rc);
}

/* --- @tx_read@, @tx_readx@ --- *
 *
 * Arguments:	@txport *tx@ = pointer to transport context
 *		@unsigned long t@ = time to wait for data (ms)
 *		@int (*filter)(const char *s, void *p)@ = filtering function
 *		@void *p@ = pointer argument for filter
 *
 * Returns:	A pointer to a line block, which must be freed using
 *		@tx_freeline@.
 *
 * Use:		Fetches a line from the buffer.  Each line is passed to the
 *		filter function in oldest-to-newest order; the filter
 *		function returns nonzero to choose a line.  If no suitable
 *		line is waiting in the raw buffer, the program blocks while
 *		more data is fetched, until the time limit @t@ is exceeded,
 *		in which case a null pointer is returned.  A null filter
 *		function is equivalent to one which always selects its line.
 */

txline *tx_readx(txport *tx, unsigned long t,
		 int (*filter)(const char *s, void *p), void *p)
{
  txline *l, **ll = &tx->ll;
  int e;
  struct timeval now, tv;
  struct timespec ts;
  unsigned f = 0;

#define f_lock 1u

  /* --- Get the time to wait until --- */

  if (t != FOREVER) {
    gettimeofday(&now, 0);
    tv_addl(&tv, &now, t / 1000, (t % 1000) * 1000);
    ts.tv_sec = tv.tv_sec;
    ts.tv_nsec = tv.tv_usec * 1000;
  }

  /* --- Check for a matching line --- */

again:
  for (; *ll; ll = &l->next) {
    l = *ll;
    if (!filter || filter(l->s, p))
      goto done;
  }
  l = 0;

  /* --- Lock the buffer --- *
   *
   * The following operations require a lock on the buffer, so we obtain that
   * here.
   */

  if (!(f & f_lock)) {
    if ((e = pthread_mutex_lock(&tx->mx)) != 0) {
      err_report(ERR_TXPORT, ERRTX_READ, e,
		 "error locking mutex: %s", strerror(errno));
      goto done;
    }
    f |= f_lock;
  }

  /* --- Push more stuff through the line buffer --- */

check:
  if (DA_LEN(&tx->buf)) {
    trace_block(1u, "incoming data", DA(&tx->buf), DA_LEN(&tx->buf));
    lbuf_snarf(&tx->lb, DA(&tx->buf), DA_LEN(&tx->buf));
    DA_SHRINK(&tx->buf, DA_LEN(&tx->buf));
    goto again;
  }

  /* --- If nothing else can arrive, give up --- */

  if (tx->s == TX_CLOSE) {
    lbuf_close(&tx->lb);
    tx->s = TX_CLOSED;
    goto again;
  }
  if (!t || tx->s == TX_CLOSED)
    goto done;
  gettimeofday(&now, 0);
  if (TV_CMP(&now, >=, &tv))
    goto done;

  /* --- Wait for some more data to arrive --- */

  if (t == FOREVER)
    e = pthread_cond_wait(&tx->cv, &tx->mx);
  else
    e = pthread_cond_timedwait(&tx->cv, &tx->mx, &ts);
  if (e && e != ETIMEDOUT && e != EINTR) {
    err_report(ERR_TXPORT, ERRTX_READ, e,
	       "error waiting on condvar: %s", strerror(errno));
    goto done;
  }
  goto check;

  /* --- Everything is finished --- */

done:
  if (f & f_lock)
    pthread_mutex_unlock(&tx->mx);
  return (l);

#undef f_lock
}

txline *tx_read(txport *tx, unsigned long t)
{
  return (tx_readx(tx, t, 0, 0));
}

/* --- @tx_freeline@ --- *
 *
 * Arguments:	@txline *l@ = pointer to line
 *
 * Returns:	---
 *
 * Use:		Frees a line block.
 */

void tx_freeline(txline *l)
{
  txport *tx = l->tx;
  if (l->next)
    l->next->prev = l->prev;
  else
    tx->ll_tail = l->prev;
  if (l->prev)
    l->prev->next = l->next;
  else
    tx->ll = l->next;
  xfree(l->s);
  DESTROY(l);
}

/* --- @tx_destroy@ --- *
 *
 * Arguments:	@txport *tx@ = transport context
 *
 * Returns:	---
 *
 * Use:		Destroys a transport.
 */

void tx_destroy(txport *tx)
{
  txline *l, *ll;

  if (tx->s == TX_READY) {
    pthread_mutex_lock(&tx->mx);
    if (tx->s == TX_READY)
      pthread_cancel(tx->tid);
    pthread_mutex_unlock(&tx->mx);
  }
  pthread_mutex_destroy(&tx->mx);
  pthread_cond_destroy(&tx->cv);
  DA_DESTROY(&tx->buf);
  lbuf_destroy(&tx->lb);
  for (l = tx->ll; l; l = ll) {
    ll = l->next;
    xfree(l->s);
    DESTROY(l);
  }
  tx->ops->destroy(tx);    
}

/*----- That's all, folks -------------------------------------------------*/
